import asyncio
from collections.abc import Generator

import aiofiles
from aiopath import AsyncPath
from asyncstdlib.functools import lru_cache

from .filesystems import FSHandlers
from .logs import logger


async def find_mount(path: AsyncPath) -> AsyncPath:
    """Find the mount point for a given path."""
    path = await path.absolute()
    while not await path.is_mount():
        path = path.parent
    return path


async def get_all_mounts() -> Generator:
    """Get a list of all mounted filesystems."""
    async with aiofiles.open("/proc/mounts", mode="r") as output:
        async for line in output:
            yield line.split()[1]


@lru_cache(maxsize=1024)
async def mount_to_fs_handler(path: AsyncPath) -> str:
    # TODO: This is a hacky way to get the filesystem type, but it works for now.
    # Maybe with libblkid Python bindings?
    logger.info(f"Getting filesystem for mount: {path}")
    proc = await asyncio.create_subprocess_exec(
        "stat", "-f", "-L", "-c", "%T", str(path), stdout=asyncio.subprocess.PIPE, stdin=asyncio.subprocess.PIPE
    )
    stdout, _ = await proc.communicate()

    if proc.returncode != 0:
        err = f"Unable to get filesystem for {path}"
        logger.error(err)
        raise RuntimeError(err)

    fs = stdout.decode().strip()
    logger.info(f"Filesystem for {path}: {fs}")

    try:
        return FSHandlers[fs]
    except KeyError:
        err = f"Unsupported filesystem: {fs}"
        logger.error(err)
        raise RuntimeError(err)
